@@ -22,33 +22,43 @@
 
 #pragma once
 
-#include <sqlite3.h>
-
 #include "medialibrary/IMedia.h"
-#include "medialibrary/filesystem/IFileSystemFactory.h"
-#include "File.h"
 #include "Thumbnail.h"
 #include "database/DatabaseHelpers.h"
-#include "medialibrary/IMetadata.h"
 #include "Metadata.h"
-
-#include <atomic>
+#include "medialibrary/IGenre.h"
 
 namespace medialibrary
 {
 
-class Album;
-class Folder;
-class ShowEpisode;
-class AlbumTrack;
-
-class Media;
-
 class Media : public IMedia,
               public DatabaseHelpers<Media>,
               public std::enable_shared_from_this<Media>
 {
     public:
+        enum class ImportType : uint8_t
+        {
+            /**
+             * Internal media, which was discovered and imported by the media
+             * library.
+             */
+            Internal,
+            /**
+             * External media, ie. media that were not discovered by the media
+             * library, but that were added manually by the user.
+             * These media are not analyzed, so their subtype, tracks, or other
+             * details are not known.
+             * They can, however, be used to store meta or be included in the
+             * playback history.
+             */
+            External,
+            /**
+             * Represent a stream, which is a specific kind of External media.
+             * This type of media is also intended to be inserted manually by
+             * the user.
+             */
+            Stream,
+        };
         struct Table
         {
             static const std::string Name;
@@ -59,51 +69,126 @@ class Media : public IMedia,
         {
             static const std::string Name;
         };
+        enum class Triggers : uint8_t
+        {
+            IsPresent,
+            CascadeFileDeletion,
+            CascadeFileUpdate,
+            IncrementNbPlaylist,
+            DecrementNbPlaylist,
+            InsertFts,
+            DeleteFts,
+            UpdateFts,
+        };
+        enum class Indexes : uint8_t
+        {
+            LastPlayedDate,
+            Presence,
+            Types,
+            //LastUsageDate, Rename in model 34
+            InsertionDate,
+            Folder,
+            MediaGroup,
+            Progress,
+            AlbumTrack,
+            Duration,
+            ReleaseDate,
+            PlayCount,
+            Title,
+            FileName,
+            GenreId,
+            ArtistId,
+        };
 
         // Those should be private, however the standard states that the expression
         // ::new (pv) T(std::forward(args)...)
         // shall be well-formed, and private constructor would prevent that.
         // There might be a way with a user-defined allocator, but we'll see that later...
         Media( MediaLibraryPtr ml , sqlite::Row& row );
+        ///
+        /// \brief Media Construct a discovered media
+        /// \param ml A media library instance pointer
+        /// \param title The media title
+        /// \param type The media type
+        /// \param duration The media duration
+        /// \param deviceId The device containing this media's ID
+        /// \param folderId The folder containing this media's ID
+        ///
         Media( MediaLibraryPtr ml, const std::string& title, Type type,
                int64_t duration, int64_t deviceId, int64_t folderId );
 
+        Media( MediaLibraryPtr ml, const std::string& fileName, ImportType type,
+               int64_t duration );
+
         static std::shared_ptr<Media> create( MediaLibraryPtr ml, Type type,
                                               int64_t deviceId, int64_t folderId,
                                               const std::string& fileName,
                                               int64_t duration );
+        static std::shared_ptr<Media> createExternal( MediaLibraryPtr ml,
+                                                      const std::string& fileName,
+                                                      int64_t duration );
+        static std::shared_ptr<Media> createStream( MediaLibraryPtr ml,
+                                                    const std::string& fileName );
         static void createTable( sqlite::Connection* connection );
-        static void createTriggers( sqlite::Connection* connection, uint32_t modelVersion );
+        static void createTriggers( sqlite::Connection* connection );
+        static void createIndexes( sqlite::Connection* connection );
         static std::string schema( const std::string& tableName, uint32_t dbModel );
+        static std::string trigger( Triggers trigger, uint32_t dbModel );
+        static std::string triggerName( Triggers trigger, uint32_t dbModel );
+        static std::string index( Indexes index, uint32_t dbModel );
+        static std::string indexName( Indexes index, uint32_t dbModel );
         static bool checkDbModel( MediaLibraryPtr ml );
 
         virtual int64_t id() const override;
         virtual Type type() const override;
-        void setType(Type type);
+        /**
+         * @brief setType Will update the media type and trigger a media refresh
+         * if the previous type was Unknown.
+         * @param type The new media type
+         * @return true if the database update was successful, false otherwise
+         */
+        bool setType( Type type ) override;
+        /**
+         * @brief setTypeInternal Will only update the media type in database
+         * @param type The new media type
+         * @return true if the database update was successful, false otherwise
+         */
+        bool setTypeInternal( Type type );
         virtual SubType subType() const override;
-        void setSubType( SubType subType );
+        bool setSubTypeUnknown();
 
         virtual const std::string& title() const override;
-        virtual bool setTitle( const std::string& title ) override;
+        bool setTitle( const std::string& title, bool forced );
         ///
-        /// \brief setTitleBuffered Mark the media as changed but doesn't save the change in DB
-        /// Querying the title after this method will return the new title, but it won't appear in DB
-        /// until save() is called
+        /// \brief setForcedTitle Force the forced_title field to true
+        /// \param mediaId The targetted media id
         ///
-        void setTitleBuffered( const std::string& title );
+        /// This is only used for migration 23 -> 24
+        ///
+        static bool setForcedTitle(MediaLibraryPtr ml, int64_t mediaId );
+        ///
+        /// \brief isForcedTitle returns true if the title is forced by the user
+        ///
+        /// This is only intended for unit tests purposes
+        ///
+        bool isForcedTitle() const;
+
         // Should only be used by 13->14 migration
         bool setFileName( std::string fileName );
-        virtual AlbumTrackPtr albumTrack() const override;
-        void setAlbumTrack( AlbumTrackPtr albumTrack );
+        bool markAsAlbumTrack( int64_t albumId, uint32_t trackNb,
+                               uint32_t discNumber, int64_t artistId, Genre* genre );
         virtual int64_t duration() const override;
-        void setDuration( int64_t duration);
+        virtual float lastPosition() const override;
+        virtual int64_t lastTime() const override;
+        bool setDuration( int64_t duration);
         virtual ShowEpisodePtr showEpisode() const override;
-        void setShowEpisode( ShowEpisodePtr episode );
+        bool setShowEpisode( ShowEpisodePtr episode );
         virtual bool addLabel( LabelPtr label ) override;
         virtual bool removeLabel( LabelPtr label ) override;
         virtual Query<ILabel> labels() const override;
         virtual uint32_t playCount() const override;
-        virtual bool increasePlayCount() override;
+        virtual ProgressResult setLastPosition( float lastPosition ) override;
+        virtual ProgressResult setLastTime( int64_t lastTime ) override;
         virtual bool setPlayCount( uint32_t playCount ) override;
         virtual time_t lastPlayedDate() const override;
         virtual bool removeFromHistory() override;
@@ -112,30 +197,41 @@ class Media : public IMedia,
         virtual const std::vector<FilePtr>& files() const override;
         virtual const std::string& fileName() const override;
         virtual MoviePtr movie() const override;
-        void setMovie( MoviePtr movie );
-        bool addVideoTrack( const std::string& codec, unsigned int width,
+        bool setMovie( MoviePtr movie );
+        bool addVideoTrack( std::string codec, unsigned int width,
                             unsigned int height, uint32_t fpsNum,
                             uint32_t fpsDen, uint32_t bitrate, uint32_t sarNum,
-                            uint32_t sarDen, const std::string& language,
-                            const std::string& description );
+                            uint32_t sarDen, std::string language,
+                            std::string description );
         virtual Query<IVideoTrack> videoTracks() const override;
-        bool addAudioTrack( const std::string& codec, unsigned int bitrate, unsigned int sampleRate,
-                            unsigned int nbChannels, const std::string& language, const std::string& desc );
+        bool addAudioTrack( std::string codec, unsigned int bitrate,
+                            unsigned int sampleRate, unsigned int nbChannels,
+                            std::string language, std::string desc,
+                            int64_t attachedFileId );
         bool addSubtitleTrack( std::string codec, std::string language,
-                               std::string description, std::string encoding );
+                               std::string description, std::string encoding,
+                               int64_t attachedFileId );
         virtual Query<IAudioTrack> audioTracks() const override;
+        /**
+         * @brief integratedAudioTracks Equivalent to audioTracks, but doesn't
+         *                              include the tracks on attached files
+         */
+        Query<IAudioTrack> integratedAudioTracks() const;
         virtual Query<ISubtitleTrack> subtitleTracks() const override;
+        /**
+         * @brief integratedSubtitleTracks Equivalent to subtitleTracks, but doesn't
+         *                                 include the tracks on external files
+         */
+        Query<ISubtitleTrack> integratedSubtitleTracks();
         virtual Query<IChapter> chapters( const QueryParameters* params ) const override;
         bool addChapter( int64_t offset, int64_t duration, std::string name );
         std::shared_ptr<Thumbnail> thumbnail( ThumbnailSizeType sizeType ) const;
         virtual const std::string& thumbnailMrl( ThumbnailSizeType sizeType ) const override;
-        virtual bool isThumbnailGenerated(  ThumbnailSizeType sizeType ) const override;
+        virtual ThumbnailStatus thumbnailStatus( ThumbnailSizeType sizeType ) const override;
         virtual bool setThumbnail( const std::string &thumbnail,
                                    ThumbnailSizeType sizeType ) override;
-        bool setThumbnail( const std::string& thumbnail, Thumbnail::Origin origin,
-                           ThumbnailSizeType sizeType, bool isOwned );
         bool setThumbnail( std::shared_ptr<Thumbnail> thumbnail );
-        void removeThumbnail( ThumbnailSizeType sizeType );
+        virtual bool removeThumbnail( ThumbnailSizeType sizeType ) override;
         virtual unsigned int insertionDate() const override;
         virtual unsigned int releaseDate() const override;
         /**
@@ -152,17 +248,28 @@ class Media : public IMedia,
         virtual bool setMetadata( MetadataType type, const std::string& value ) override;
         virtual bool setMetadata( MetadataType type, int64_t value ) override;
         virtual bool unsetMetadata( MetadataType type ) override;
-        virtual bool setMetadata( std::unordered_map<MetadataType, std::string> metadata ) override;
+        virtual bool setMetadata( const std::unordered_map<MetadataType, std::string>& metadata ) override;
 
         virtual bool requestThumbnail( ThumbnailSizeType sizeType, uint32_t desiredWidth,
                                        uint32_t desiredHeight, float position ) override;
+        virtual bool isDiscoveredMedia() const override;
+        virtual bool isExternalMedia() const override;
+        virtual bool isStream() const override;
 
-        void setReleaseDate( unsigned int date );
+        virtual bool addToGroup( IMediaGroup& group ) override;
+        virtual bool addToGroup( int64_t groupId ) override;
+        void setMediaGroupId( int64_t groupId );
+        virtual bool removeFromGroup() override;
+        virtual MediaGroupPtr group() const override;
+        virtual int64_t groupId() const override;
+        virtual bool regroup() override;
+
+        bool setReleaseDate( unsigned int date );
         int64_t deviceId() const; // Used for unit tests purposes only
-        void setDeviceId( int64_t deviceId );
         int64_t folderId() const; // Used for unit tests purposes only
-        void setFolderId( int64_t folderId );
-        bool save();
+        bool markAsInternal( Type type, int64_t duration, int64_t deviceId,
+                             int64_t folderId );
+        bool convertToExternal();
 
         std::shared_ptr<File> addFile( const fs::IFile& fileFs, int64_t parentFolderId,
                                        bool isFolderFsRemovable, IFile::Type type );
@@ -176,7 +283,24 @@ class Media : public IMedia,
         virtual bool removeBookmark( int64_t time ) override;
         virtual bool removeAllBookmarks() override;
 
-        static Query<IMedia> listAll(MediaLibraryPtr ml, Type type, const QueryParameters* params );
+        virtual bool isPresent() const override;
+
+        virtual ArtistPtr artist() const override;
+        virtual int64_t artistId() const override;
+        virtual GenrePtr genre() const override;
+        virtual int64_t genreId() const override;
+        virtual unsigned int trackNumber() const override;
+        virtual AlbumPtr album() const override;
+        virtual int64_t albumId() const override;
+        virtual uint32_t discNumber() const override;
+        virtual const std::string& lyrics() const override;
+        bool setLyrics( std::string lyrics );
+
+        static Query<IMedia> listAll( MediaLibraryPtr ml, Type type,
+                                     const QueryParameters* params,
+                                     Media::SubType subType );
+        static Query<IMedia> listInProgress( MediaLibraryPtr ml, Type type,
+                                             const QueryParameters* params );
 
         static Query<IMedia> search( MediaLibraryPtr ml, const std::string& title,
                                      const QueryParameters* params );
@@ -202,13 +326,14 @@ class Media : public IMedia,
                                                  const std::string& pattern,
                                                  Type type, int64_t folderId,
                                                  const QueryParameters* params );
-        static Query<IMedia> fromVideoGroup( MediaLibraryPtr ml,
-                                             const std::string& name,
+        static Query<IMedia> fromMediaGroup( MediaLibraryPtr ml,
+                                             int64_t groupId, Type type,
                                              const QueryParameters* params );
-        static Query<IMedia> searchFromVideoGroup( MediaLibraryPtr ml,
-                                                   const std::string& groupName,
+        static Query<IMedia> searchFromMediaGroup( MediaLibraryPtr ml,
+                                                   int64_t groupId, Type type,
                                                    const std::string& pattern,
                                                    const QueryParameters* params );
+        static bool setMediaGroup(MediaLibraryPtr ml, int64_t mediaId, int64_t groupId );
 
         static bool clearHistory( MediaLibraryPtr ml );
         static bool removeOldMedia( MediaLibraryPtr ml, std::chrono::seconds maxLifeTime );
@@ -223,12 +348,50 @@ class Media : public IMedia,
          */
         static bool resetSubTypes( MediaLibraryPtr ml );
 
+        static bool regroupAll( MediaLibraryPtr ml );
+
+        static Query<IMedia> tracksFromGenre( MediaLibraryPtr ml, int64_t genreId,
+                                              IGenre::TracksIncluded included,
+                                              const QueryParameters* params );
+
 private:
-        static std::string addRequestJoin(const QueryParameters* params, bool forceFile , bool forceAlbumTrack);
+        enum class PositionTypes : uint8_t
+        {
+            /**
+             * The provided position is at the begining of the media, it should
+             * be stored as -1 and the playback will not be considered started
+            */
+            Begin,
+            /**
+             * The provided position is at the end of the media, it should be stored
+             * as -1 to indicate that the media has been played fully and the
+             * playcount should be updated.
+             */
+            End,
+            /**
+             * The position isn't anything special and should be stored as is
+             */
+            Any,
+        };
+        static std::string addRequestJoin(const QueryParameters* params, bool forceFile );
         static std::string sortRequest( const QueryParameters* params );
         static Query<IMedia> fetchHistoryByType( MediaLibraryPtr ml, IMedia::Type type );
-        bool shouldUpdateThumbnail( Thumbnail& currentThumbnail,
-                                    Thumbnail::Origin newOrigin );
+        static bool shouldUpdateThumbnail( const Thumbnail& currentThumbnail );
+        static std::shared_ptr<Media> createExternalMedia( MediaLibraryPtr ml,
+                                                           const std::string& mrl,
+                                                           ImportType importType,
+                                                           int64_t duration );
+        std::vector<std::shared_ptr<Media>> fetchMatchingUngrouped();
+        PositionTypes computePositionType( float position ) const;
+        ProgressResult setLastPositionAndTime( PositionTypes positionType,
+                                               float lastPos, int64_t lastTime );
+
+        /*
+         * Marked as private so that it can only be used through IMedia interface
+         * Internally, we have to use the setTitle( title, force ) overload
+        */
+        virtual bool setTitle( const std::string& title ) override;
+
 
 private:
         MediaLibraryPtr m_ml;
@@ -238,27 +401,37 @@ private:
         Type m_type;
         SubType m_subType;
         int64_t m_duration;
+        float m_lastPosition;
+        int64_t m_lastTime;
         unsigned int m_playCount;
-        std::time_t m_lastPlayedDate;
-        const std::time_t m_insertionDate;
+        time_t m_lastPlayedDate;
+        const time_t m_insertionDate;
         unsigned int m_releaseDate;
         std::string m_title;
         // We store the filename as a shortcut when sorting. The filename (*not* the title
         // might be used as a fallback
         std::string m_filename;
         bool m_isFavorite;
+        bool m_isPresent;
         int64_t m_deviceId;
         uint32_t m_nbPlaylists;
         int64_t m_folderId;
+        ImportType m_importType;
+        int64_t m_groupId;
+        bool m_forcedTitle;
+        int64_t m_artistId;
+        int64_t m_genreId;
+        uint32_t m_trackNumber;
+        int64_t m_albumId;
+        uint32_t m_discNumber;
+        std::string m_lyrics;
 
         // Auto fetched related properties
-        mutable AlbumTrackPtr m_albumTrack;
         mutable ShowEpisodePtr m_showEpisode;
         mutable MoviePtr m_movie;
         mutable std::vector<FilePtr> m_files;
         mutable Metadata m_metadata;
         mutable std::shared_ptr<Thumbnail> m_thumbnails[Thumbnail::SizeToInt( ThumbnailSizeType::Count )];
-        bool m_changed;
 
         friend Media::Table;
 };
