@@ -26,17 +26,17 @@
 
 #if WINAPI_FAMILY_PARTITION (WINAPI_PARTITION_DESKTOP)
 
-#include "medialibrary/IDeviceLister.h"
+#include "filesystem/common/CommonDeviceLister.h"
 
 namespace medialibrary
 {
 namespace fs
 {
 
-class DeviceLister : public IDeviceLister
+class DeviceLister : public CommonDeviceLister
 {
 public:
-    virtual std::vector<std::tuple<std::string, std::string, bool>> devices() const override;
+    virtual std::vector<Device> devices() const override;
 };
 
 }
