@@ -39,6 +39,7 @@
 #include "utils/Filename.h"
 #include "utils/Url.h"
 #include "utils/Strings.h"
+#include "utils/Enums.h"
 
 #include <algorithm>
 
@@ -54,18 +55,20 @@ int64_t parser::Task::* const Task::Table::PrimaryKey = &parser::Task::m_id;
 
 Task::Task( MediaLibraryPtr ml, sqlite::Row& row )
     : m_ml( ml )
+    , m_id( row.extract<decltype(m_id)>() )
+    , m_step( row.extract<decltype(m_step)>() )
+    , m_attemptsRemaining( row.extract<decltype(m_attemptsRemaining)>() )
+    , m_type( row.extract<decltype(m_type)>() )
+    , m_mrl( row.extract<decltype(m_mrl)>() )
+    , m_fileType( row.extract<decltype(m_fileType)>() )
+    , m_fileId( row.extract<decltype(m_fileId)>() )
+    , m_parentFolderId( row.extract<decltype(m_parentFolderId)>() )
+    , m_linkToId( row.extract<decltype(m_linkToId)>() )
+    , m_linkToType( row.extract<decltype(m_linkToType)>() )
+    , m_linkExtra( row.extract<decltype(m_linkExtra)>() )
+    , m_linkToMrl( row.extract<decltype(m_linkToMrl)>() )
 {
-    row >> m_id
-        >> m_step
-        >> m_retryCount
-        >> m_type
-        >> m_mrl
-        >> m_fileType
-        >> m_fileId
-        >> m_parentFolderId
-        >> m_linkToId
-        >> m_linkToType
-        >> m_linkExtra;
+    assert( row.hasRemainingColumns() == false );
 }
 
 Task::Task( MediaLibraryPtr ml, std::string mrl, std::shared_ptr<fs::IFile> fileFs,
@@ -73,6 +76,7 @@ Task::Task( MediaLibraryPtr ml, std::string mrl, std::shared_ptr<fs::IFile> file
             std::shared_ptr<fs::IDirectory> parentFolderFs,
             IFile::Type fileType )
     : m_ml( ml )
+    , m_attemptsRemaining( Settings::MaxTaskAttempts )
     , m_type( Type::Creation )
     , m_mrl( std::move( mrl ) )
     , m_fileType( fileType )
@@ -87,10 +91,12 @@ Task::Task( MediaLibraryPtr ml, std::shared_ptr<File> file,
             std::shared_ptr<fs::IFile> fileFs, std::shared_ptr<Folder> parentFolder,
             std::shared_ptr<fs::IDirectory> parentFolderFs )
     : m_ml( ml )
+    , m_attemptsRemaining( Settings::MaxTaskAttempts )
     , m_type( Type::Refresh )
     , m_mrl( file->mrl() )
     , m_fileType( file->type() )
     , m_fileId( file->id() )
+    , m_parentFolderId( parentFolder->id() )
     , m_file( std::move( file ) )
     , m_fileFs( std::move( fileFs ) )
     , m_parentFolder( std::move( parentFolder ) )
@@ -101,6 +107,7 @@ Task::Task( MediaLibraryPtr ml, std::shared_ptr<File> file,
 Task::Task( MediaLibraryPtr ml, std::string mrl, int64_t linkToId,
             Task::LinkType linkToType, int64_t linkExtra )
     : m_ml( ml )
+    , m_attemptsRemaining( Settings::MaxLinkTaskAttempts )
     , m_type( Type::Link )
     , m_mrl( std::move( mrl ) )
     , m_linkToId( linkToId )
@@ -109,19 +116,25 @@ Task::Task( MediaLibraryPtr ml, std::string mrl, int64_t linkToId,
 {
 }
 
-Task::Task( MediaLibraryPtr ml, std::string mrl )
+Task::Task( MediaLibraryPtr ml, std::string mrl, IFile::Type fileType,
+            std::string linkToMrl, IItem::LinkType linkToType, int64_t linkExtra )
     : m_ml( ml )
-    , m_type( Type::Restore )
+    , m_attemptsRemaining( Settings::MaxLinkTaskAttempts )
+    , m_type( Type::Link )
     , m_mrl( std::move( mrl ) )
+    , m_fileType( fileType )
+    , m_linkToType( linkToType )
+    , m_linkExtra( linkExtra )
+    , m_linkToMrl( std::move( linkToMrl ) )
 {
 }
 
-Task::Task( std::string mrl, IFile::Type fileType, unsigned int playlistIndex )
-    : m_type( Type::Temporary )
+Task::Task( MediaLibraryPtr ml, std::string mrl , IFile::Type fileType )
+    : m_ml( ml )
+    , m_attemptsRemaining( Settings::MaxTaskAttempts )
+    , m_type( Type::Restore )
     , m_mrl( std::move( mrl ) )
     , m_fileType( fileType )
-    , m_linkToType( LinkType::Playlist )
-    , m_linkExtra( playlistIndex )
 {
 }
 
@@ -133,21 +146,28 @@ void Task::markStepCompleted( Step stepCompleted )
 
 bool Task::saveParserStep()
 {
-    static const std::string req = "UPDATE " + Task::Table::Name + " SET step = ?, "
-            "retry_count = 0 WHERE id_task = ?";
-    if ( sqlite::Tools::executeUpdate( m_ml->getConn(), req, m_step, m_id ) == false )
+    static const std::string req = "UPDATE " + Table::Name + " SET step = ?, "
+            "attempts_left = "
+            "(CASE type "
+                "WHEN " + utils::enum_to_string( Type::Link ) + " "
+                "THEN (SELECT max_link_task_attempts FROM Settings) "
+                "ELSE (SELECT max_task_attempts FROM Settings) END) "
+            "WHERE id_task = ?";
+    if ( sqlite::Tools::executeUpdate( m_ml->getConn(), req, m_step,
+                                       m_id ) == false )
         return false;
-    m_retryCount = 0;
+    m_attemptsRemaining = m_type != Type::Link ?
+                Settings::MaxTaskAttempts : Settings::MaxLinkTaskAttempts;
     return true;
 }
 
 bool Task::decrementRetryCount()
 {
-    static const std::string req = "UPDATE " + Task::Table::Name + " SET "
-            "retry_count = retry_count - 1 WHERE id_task = ?";
+    static const std::string req = "UPDATE " + Table::Name + " SET "
+            "attempts_left = attempts_left + 1 WHERE id_task = ?";
     if ( sqlite::Tools::executeUpdate( m_ml->getConn(), req, m_id ) == false )
         return false;
-    --m_retryCount;
+    ++m_attemptsRemaining;
     return true;
 }
 
@@ -163,20 +183,42 @@ bool Task::isStepCompleted( Step step ) const
 {
     if ( m_type == Type::Link )
     {
-        // We always want to return true for any step which isn't the linking step
-        // since we don't want to run any service other than the linking one
-        return step != Step::Linking;
+        switch ( step )
+        {
+            case Step::MetadataExtraction:
+            {
+                if ( m_linkToType == LinkType::Media &&
+                       m_fileType == IFile::Type::Soundtrack )
+                {
+                    /* If the attached file is an audio track, we want to process
+                     * it if we haven't already, so we fallback to the default case
+                     * to run it only when the step hasn't been executed yet.
+                     * In other cases, we hardcode the results
+                     */
+                    break;
+                }
+                return true;
+            }
+            case Step::MetadataAnalysis:
+                /* We don't have anything to extract, just link tracks together */
+                return true;
+            case Step::Linking:
+                /* We definitely want to run the link step for link tasks */
+                return false;
+            default:
+                assert( !"Invalid state for linking task" );
+        }
     }
     return ( static_cast<uint8_t>( m_step ) & static_cast<uint8_t>( step ) ) != 0;
 }
 
 void Task::startParserStep()
 {
-    static const std::string req = "UPDATE " + Task::Table::Name + " SET "
-            "retry_count = retry_count + 1 WHERE id_task = ?";
+    static const std::string req = "UPDATE " + Table::Name + " SET "
+            "attempts_left = attempts_left - 1 WHERE id_task = ?";
     if ( sqlite::Tools::executeUpdate( m_ml->getConn(), req, m_id ) == false )
         return;
-    ++m_retryCount;
+    --m_attemptsRemaining;
 }
 
 uint32_t Task::goToNextService()
@@ -189,9 +231,9 @@ void Task::resetCurrentService()
     m_currentService = 0;
 }
 
-unsigned int Task::retryCount() const
+unsigned int Task::attemptsRemaining() const
 {
-    return m_retryCount;
+    return m_attemptsRemaining;
 }
 
 int64_t Task::id() const
@@ -199,11 +241,10 @@ int64_t Task::id() const
     return m_id;
 }
 
-bool Task::restoreLinkedEntities()
+bool Task::restoreLinkedEntities( LastTaskInfo& lastTask )
 {
     // No need to restore anything for link tasks, they only contain mrls & ids.
-    if ( isLinkTask() == true || isRestore() == true )
-        return true;
+    assert( needEntityRestoration() == true );
     LOG_DEBUG("Restoring linked entities of task ", m_id);
     // MRL will be empty if the task has been resumed from unparsed files
     // (during 11 -> 12 migration)
@@ -275,6 +316,7 @@ bool Task::restoreLinkedEntities()
                 LOG_INFO( "Duplicated task after mrl update, discarding the "
                           "duplicate." );
                 destroy( m_ml, m_id );
+                return false;
             }
         }
     }
@@ -282,21 +324,30 @@ bool Task::restoreLinkedEntities()
     // Now we always have a valid MRL, but we might not have a fileId
     // In any case, we need to fetch the corresponding FS entities
 
-    auto fsFactory = m_ml->fsFactoryForMrl( mrl );
-    if ( fsFactory == nullptr )
+    if ( lastTask.parentFolderId == m_parentFolderId && m_parentFolderId != 0 )
     {
-        LOG_WARN( "No fs factory matched the task mrl (", mrl, "). Postponing" );
-        return false;
+        m_parentFolderFs = lastTask.fsDir;
     }
+    else
+    {
+        auto fsFactory = m_ml->fsFactoryForMrl( mrl );
+        if ( fsFactory == nullptr )
+        {
+            LOG_WARN( "No fs factory matched the task mrl (", mrl, "). Postponing" );
+            return false;
+        }
 
-    try
-    {
-        m_parentFolderFs = fsFactory->createDirectory( utils::file::directory( mrl ) );
-    }
-    catch ( const fs::errors::System& ex )
-    {
-        LOG_ERROR( "Failed to restore task: ", ex.what() );
-        return false;
+        try
+        {
+            m_parentFolderFs = fsFactory->createDirectory( utils::file::directory( mrl ) );
+        }
+        catch ( const fs::errors::System& ex )
+        {
+            LOG_ERROR( "Failed to restore task: ", ex.what() );
+            return false;
+        }
+        lastTask.fsDir = m_parentFolderFs;
+        lastTask.parentFolderId = m_parentFolderId;
     }
 
     try
@@ -360,41 +411,52 @@ bool Task::restoreLinkedEntities()
     return true;
 }
 
+bool Task::setMrl( MediaLibraryPtr ml, int64_t taskId, const std::string& mrl )
+{
+    static const std::string req = "UPDATE " + Table::Name + " SET "
+            "mrl = ? WHERE id_task = ?";
+    return sqlite::Tools::executeUpdate( ml->getConn(), req, mrl, taskId );
+}
+
 void Task::setMrl( std::string newMrl )
 {
     if ( m_mrl == newMrl )
         return;
-    static const std::string req = "UPDATE " + Task::Table::Name + " SET "
-            "mrl = ? WHERE id_task = ?";
-    if ( sqlite::Tools::executeUpdate( m_ml->getConn(), req, newMrl, m_id ) == false )
+    if ( setMrl( m_ml, m_id, newMrl ) == false )
         return;
     m_mrl = std::move( newMrl );
 }
 
-void Task::createTable( sqlite::Connection* dbConnection, uint32_t dbModel )
+void Task::createTable( sqlite::Connection* dbConnection )
 {
     sqlite::Tools::executeRequest( dbConnection,
-                                   schema( Table::Name, dbModel, false ) );
+                                   schema( Table::Name, Settings::DbModelVersion ) );
 }
 
-void Task::createTriggers( sqlite::Connection* dbConnection, uint32_t dbModel )
+void Task::createTriggers( sqlite::Connection* dbConnection )
 {
-    if ( dbModel < 18 )
-        return;
-    std::string reqs[] = {
-        #include "database/tables/Task_triggers_v18.sql"
-    };
-    for ( const auto& req : reqs )
-        sqlite::Tools::executeRequest( dbConnection, req );
+    sqlite::Tools::executeRequest( dbConnection,
+                                   trigger( Triggers::DeletePlaylistLinkingTask,
+                                            Settings::DbModelVersion ) );
 }
 
-std::string Task::schema( const std::string& tableName, uint32_t dbModel,
-                          bool backup )
+void Task::createIndex( sqlite::Connection* dbConnection )
 {
+    sqlite::Tools::executeRequest( dbConnection,
+                                   index( Indexes::ParentFolderId,
+                                          Settings::DbModelVersion ) );
+    sqlite::Tools::executeRequest( dbConnection,
+                                   index( Indexes::FileId,
+                                          Settings::DbModelVersion ) );
+}
+
+std::string Task::schema( const std::string& tableName, uint32_t dbModel )
+{
+    UNUSED_IN_RELEASE( tableName );
+
     assert( tableName == Table::Name );
     if ( dbModel <= 17 )
     {
-        assert( backup == false );
         return "CREATE TABLE " + Table::Name +
         "("
             "id_task INTEGER PRIMARY KEY AUTOINCREMENT,"
@@ -409,119 +471,271 @@ std::string Task::schema( const std::string& tableName, uint32_t dbModel,
             "is_refresh BOOLEAN NOT NULL DEFAULT 0,"
             "UNIQUE(mrl, parent_playlist_id, is_refresh) ON CONFLICT FAIL,"
             "FOREIGN KEY(parent_folder_id) REFERENCES " + Folder::Table::Name
-            + "(id_folder) ON DELETE CASCADE,"
+                + "(id_folder) ON DELETE CASCADE,"
             "FOREIGN KEY(file_id) REFERENCES " + File::Table::Name
-            + "(id_file) ON DELETE CASCADE,"
+                + "(id_file) ON DELETE CASCADE,"
             "FOREIGN KEY(parent_playlist_id) REFERENCES " + Playlist::Table::Name
-            + "(id_playlist) ON DELETE CASCADE"
+                + "(id_playlist) ON DELETE CASCADE"
         ")";
     }
-    auto req = std::string{ "CREATE " } + (backup ? "TEMPORARY " : "") +
-            "TABLE " + Table::Name + (backup ? "_backup" : "" ) +
+    if ( dbModel < 20 )
+    {
+        return "CREATE TABLE " + Table::Name +
+        "("
+            "id_task INTEGER PRIMARY KEY AUTOINCREMENT,"
+            "step INTEGER NOT NULL DEFAULT 0,"
+            "retry_count INTEGER NOT NULL DEFAULT 0,"
+            "type INTEGER NOT NULL,"
+            "mrl TEXT,"
+            "file_type INTEGER NOT NULL,"
+            "file_id UNSIGNED INTEGER,"
+            "parent_folder_id UNSIGNED INTEGER,"
+            "link_to_id UNSIGNED INTEGER,"
+            "link_to_type UNSIGNED INTEGER,"
+            "link_extra UNSIGNED INTEGER,"
+            "UNIQUE(mrl,type) ON CONFLICT FAIL,"
+            "FOREIGN KEY(parent_folder_id) REFERENCES " + Folder::Table::Name
+                + "(id_folder) ON DELETE CASCADE,"
+            "FOREIGN KEY(file_id) REFERENCES " + File::Table::Name
+                + "(id_file) ON DELETE CASCADE"
+        ")";
+    }
+    if ( dbModel < 22 )
+    {
+        return "CREATE TABLE " + Table::Name +
+        "("
+            "id_task INTEGER PRIMARY KEY AUTOINCREMENT,"
+            "step INTEGER NOT NULL DEFAULT 0,"
+            "retry_count INTEGER NOT NULL DEFAULT 0,"
+            "type INTEGER NOT NULL,"
+            "mrl TEXT,"
+            "file_type INTEGER NOT NULL,"
+            "file_id UNSIGNED INTEGER,"
+            "parent_folder_id UNSIGNED INTEGER,"
+            "link_to_id UNSIGNED INTEGER NOT NULL,"
+            "link_to_type UNSIGNED INTEGER,"
+            "link_extra UNSIGNED INTEGER,"
+            "UNIQUE(mrl,type, link_to_id) ON CONFLICT FAIL,"
+            "FOREIGN KEY(parent_folder_id) REFERENCES " + Folder::Table::Name
+                + "(id_folder) ON DELETE CASCADE,"
+            "FOREIGN KEY(file_id) REFERENCES " + File::Table::Name
+                + "(id_file) ON DELETE CASCADE"
+        ")";
+    }
+    if ( dbModel < 25 )
+    {
+        return "CREATE TABLE " + Table::Name +
+        "("
+            "id_task INTEGER PRIMARY KEY AUTOINCREMENT,"
+            "step INTEGER NOT NULL DEFAULT 0,"
+            "retry_count INTEGER NOT NULL DEFAULT 0,"
+            "type INTEGER NOT NULL,"
+            "mrl TEXT,"
+            "file_type INTEGER NOT NULL,"
+            "file_id UNSIGNED INTEGER,"
+            "parent_folder_id UNSIGNED INTEGER,"
+            "link_to_id UNSIGNED INTEGER NOT NULL,"
+            "link_to_type UNSIGNED INTEGER NOT NULL,"
+            "link_extra UNSIGNED INTEGER NOT NULL,"
+            "UNIQUE(mrl,type, link_to_id, link_to_type, link_extra) "
+                "ON CONFLICT FAIL,"
+            "FOREIGN KEY(parent_folder_id) REFERENCES " + Folder::Table::Name
+                + "(id_folder) ON DELETE CASCADE,"
+            "FOREIGN KEY(file_id) REFERENCES " + File::Table::Name
+                + "(id_file) ON DELETE CASCADE"
+        ")";
+    }
+    if ( dbModel < 27 )
+    {
+        return "CREATE TABLE " + Table::Name +
+        "("
+            "id_task INTEGER PRIMARY KEY AUTOINCREMENT,"
+            "step INTEGER NOT NULL DEFAULT 0,"
+            "retry_count INTEGER NOT NULL DEFAULT 0,"
+            "type INTEGER NOT NULL,"
+            "mrl TEXT,"
+            "file_type INTEGER NOT NULL,"
+            "file_id UNSIGNED INTEGER,"
+            "parent_folder_id UNSIGNED INTEGER,"
+            "link_to_id UNSIGNED INTEGER NOT NULL,"
+            "link_to_type UNSIGNED INTEGER NOT NULL,"
+            "link_extra UNSIGNED INTEGER NOT NULL,"
+            "link_to_mrl TEXT NOT NULL,"
+            "UNIQUE(mrl,type, link_to_id, link_to_type, link_extra, link_to_mrl) "
+                "ON CONFLICT FAIL,"
+            "FOREIGN KEY(parent_folder_id) REFERENCES " + Folder::Table::Name
+                + "(id_folder) ON DELETE CASCADE,"
+            "FOREIGN KEY(file_id) REFERENCES " + File::Table::Name
+                + "(id_file) ON DELETE CASCADE"
+        ")";
+    }
+    return "CREATE TABLE " + Table::Name +
     "("
         "id_task INTEGER PRIMARY KEY AUTOINCREMENT,"
         "step INTEGER NOT NULL DEFAULT 0,"
-        "retry_count INTEGER NOT NULL DEFAULT 0,"
+        "attempts_left INTEGER NOT NULL,"
         "type INTEGER NOT NULL,"
         "mrl TEXT,"
         "file_type INTEGER NOT NULL,"
         "file_id UNSIGNED INTEGER,"
-        "parent_folder_id UNSIGNED INTEGER,";
-
-        // For linking purposes
-        // link_to_id needs to be not null, so the unique constraint always take
-        // its value into account
-        // Starting from model 22, this is true for all link_ fields.
-
-    if ( dbModel >= 22 )
-    {
-        req += "link_to_id UNSIGNED INTEGER NOT NULL,"
-               "link_to_type UNSIGNED INTEGER NOT NULL,"
-               "link_extra UNSIGNED INTEGER NOT NULL,"
-               "UNIQUE(mrl,type, link_to_id, link_to_type, link_extra) ON CONFLICT FAIL,";
-    }
-    else if ( dbModel >= 20 )
-    {
-        req += "link_to_id UNSIGNED INTEGER NOT NULL,"
-               "link_to_type UNSIGNED INTEGER,"
-               "link_extra UNSIGNED INTEGER,"
-               "UNIQUE(mrl,type, link_to_id) ON CONFLICT FAIL,";
-    }
-    else
-    {
-        req += "link_to_id UNSIGNED INTEGER, "
-               "link_to_type UNSIGNED INTEGER,"
-               "link_extra UNSIGNED INTEGER,"
-               "UNIQUE(mrl,type) ON CONFLICT FAIL,";
-    }
-    req += "FOREIGN KEY(parent_folder_id) REFERENCES " + Folder::Table::Name
+        "parent_folder_id UNSIGNED INTEGER,"
+        "link_to_id UNSIGNED INTEGER NOT NULL,"
+        "link_to_type UNSIGNED INTEGER NOT NULL,"
+        "link_extra UNSIGNED INTEGER NOT NULL,"
+        "link_to_mrl TEXT NOT NULL,"
+        "UNIQUE(mrl,type, link_to_id, link_to_type, link_extra, link_to_mrl) "
+            "ON CONFLICT FAIL,"
+        "FOREIGN KEY(parent_folder_id) REFERENCES " + Folder::Table::Name
             + "(id_folder) ON DELETE CASCADE,"
-            "FOREIGN KEY(file_id) REFERENCES " + File::Table::Name
+        "FOREIGN KEY(file_id) REFERENCES " + File::Table::Name
             + "(id_file) ON DELETE CASCADE"
     ")";
-    return req;
+}
+
+std::string Task::trigger(Task::Triggers trigger, uint32_t dbModel )
+{
+    assert( trigger == Triggers::DeletePlaylistLinkingTask );
+    assert( dbModel >= 18 );
+    return "CREATE TRIGGER " + triggerName( trigger, dbModel ) +
+           " AFTER DELETE ON " + Playlist::Table::Name +
+           " BEGIN "
+               "DELETE FROM " + Table::Name + " "
+                   "WHERE link_to_type = " +
+                           utils::enum_to_string( IItem::LinkType::Playlist ) + " "
+                       "AND link_to_id = old.id_playlist "
+                       "AND type = " +
+                           utils::enum_to_string( Type::Link ) + ";"
+           "END";
+
+}
+
+std::string Task::triggerName( Triggers trigger, uint32_t dbModel )
+{
+    UNUSED_IN_RELEASE( trigger );
+    UNUSED_IN_RELEASE( dbModel );
+
+    assert( trigger == Triggers::DeletePlaylistLinkingTask );
+    assert( dbModel >= 18 );
+    return "delete_playlist_linking_tasks";
+}
+
+std::string Task::index( Task::Indexes index, uint32_t dbModel )
+{
+    switch ( index )
+    {
+        case Indexes::ParentFolderId:
+            assert( dbModel >= 24 );
+            return "CREATE INDEX " + indexName( index, dbModel ) +
+                    " ON " + Table::Name + "(parent_folder_id)";
+        case Indexes::FileId:
+            assert( dbModel >= 34 );
+            return "CREATE INDEX " + indexName( index, dbModel ) +
+                    " ON " + Table::Name + "(file_id)";
+    }
+    return "<invalid request>";
+}
+
+std::string Task::indexName( Task::Indexes index, uint32_t dbModel )
+{
+    UNUSED_IN_RELEASE( dbModel );
+
+    switch ( index )
+    {
+        case Indexes::ParentFolderId:
+            assert( dbModel >= 24 );
+            return "task_parent_folder_id_idx";
+        case Indexes::FileId:
+            assert( dbModel >= 34 );
+            return "task_file_id_idx";
+    }
+    return "<invalid request>";
 }
 
 bool Task::checkDbModel( MediaLibraryPtr ml )
 {
-    return sqlite::Tools::checkSchema( ml->getConn(),
-                                       schema( Table::Name, Settings::DbModelVersion, false ),
-                                       Table::Name );
+    OPEN_READ_CONTEXT( ctx, ml->getConn() );
+
+    return sqlite::Tools::checkTableSchema(
+                                       schema( Table::Name, Settings::DbModelVersion ),
+                                       Table::Name ) &&
+           sqlite::Tools::checkTriggerStatement(
+                trigger( Triggers::DeletePlaylistLinkingTask, Settings::DbModelVersion ),
+                triggerName( Triggers::DeletePlaylistLinkingTask, Settings::DbModelVersion ) ) &&
+           sqlite::Tools::checkIndexStatement(
+                index( Indexes::ParentFolderId, Settings::DbModelVersion ),
+                indexName( Indexes::ParentFolderId, Settings::DbModelVersion ) );
 }
 
 bool Task::resetRetryCount( MediaLibraryPtr ml )
 {
-    static const std::string req = "UPDATE " + Task::Table::Name + " SET "
-            "retry_count = 0 WHERE step & ?1 != ?1 AND step != ?2";
-    return sqlite::Tools::executeUpdate( ml->getConn(), req, Step::Completed,
-                                         Step::Linking );
+    auto t = ml->getConn()->newTransaction();
+    static const std::string req = "UPDATE " + Table::Name + " SET "
+            "attempts_left = (SELECT max_task_attempts FROM SETTINGS) "
+            "WHERE step & ?1 != ?1 AND type = ?";
+    static const std::string linkReq = "UPDATE " + Table::Name + " SET "
+            "attempts_left = (SELECT max_link_task_attempts FROM SETTINGS) "
+            "WHERE step & ?1 != ?1 AND type = ?";
+    auto res = sqlite::Tools::executeUpdate( ml->getConn(), req, Step::Completed,
+                                             Type::Creation ) &&
+               sqlite::Tools::executeUpdate( ml->getConn(), linkReq, Step::Completed,
+                                             Type::Link );
+    if ( res == false )
+        return false;
+    t->commit();
+    return true;
 }
 
 bool Task::resetParsing( MediaLibraryPtr ml )
 {
-    assert( sqlite::Transaction::transactionInProgress() == true );
-    static const std::string resetReq = "UPDATE " + Task::Table::Name + " SET "
-            "retry_count = 0, step = ?";
+    assert( sqlite::Transaction::isInProgress() == true );
+    static const std::string resetReq = "UPDATE " + Table::Name + " SET "
+            "attempts_left = (SELECT max_task_attempts FROM Settings), "
+            "step = ? WHERE type != ?";
+    static const std::string resetLinkReq = "UPDATE " + Table::Name + " SET "
+            "attempts_left = (SELECT max_link_task_attempts FROM Settings), "
+            "step = ? WHERE type = ?";
     /* We also want to delete the refresh tasks, since we are going to rescan
      * all existing media anyway
      */
-    static const std::string deleteRefreshReq = "DELETE FROM " + Task::Table::Name +
+    static const std::string deleteRefreshReq = "DELETE FROM " + Table::Name +
             " WHERE type = ?";
-    return sqlite::Tools::executeUpdate( ml->getConn(), resetReq, Step::None ) &&
-           sqlite::Tools::executeDelete( ml->getConn(), deleteRefreshReq, Type::Refresh );
+    return sqlite::Tools::executeDelete( ml->getConn(), deleteRefreshReq, Type::Refresh ) &&
+            sqlite::Tools::executeUpdate( ml->getConn(), resetReq,
+                                          Step::None, Type::Link ) &&
+            sqlite::Tools::executeUpdate( ml->getConn(), resetLinkReq,
+                                          Step::None, Type::Link );
 }
 
 std::vector<std::shared_ptr<Task>> Task::fetchUncompleted( MediaLibraryPtr ml )
 {
-    static const std::string req = "SELECT * FROM " + Task::Table::Name + " t"
-        " LEFT JOIN " + File::Table::Name + " f ON f.id_file = t.file_id"
-        " LEFT JOIN " + Folder::Table::Name + " fol ON f.folder_id = fol.id_folder"
+    static const std::string req = "SELECT t.* FROM " + Table::Name + " t"
+        " LEFT JOIN " + Folder::Table::Name + " fol ON t.parent_folder_id = fol.id_folder"
         " LEFT JOIN " + Device::Table::Name + " d ON d.id_device = fol.device_id"
-        " WHERE step & ? != ? AND retry_count <= ? AND (d.is_present != 0 OR "
-        " t.file_id IS NULL)";
+        " WHERE step & ? != ? AND attempts_left > 0 AND "
+            "(d.is_present != 0 OR (t.parent_folder_id IS NULL AND t.type = ?))"
+        " ORDER BY parent_folder_id";
     return Task::fetchAll<Task>( ml, req, Step::Completed,
-                                 Step::Completed, parser::MaxNbRetries );
+                                 Step::Completed, Type::Link );
 }
 
 std::shared_ptr<Task>
-Task::create( MediaLibraryPtr ml, std::string mrl, std::shared_ptr<fs::IFile> fileFs,
+Task::create( MediaLibraryPtr ml, std::shared_ptr<fs::IFile> fileFs,
               std::shared_ptr<Folder> parentFolder, std::shared_ptr<fs::IDirectory> parentFolderFs,
               IFile::Type fileType )
 {
     auto parentFolderId = parentFolder->id();
 
-    std::shared_ptr<Task> self = std::make_shared<Task>( ml, std::move( mrl ), std::move( fileFs ),
-        std::move( parentFolder ), std::move( parentFolderFs ), fileType );
-    const std::string req = "INSERT INTO " + Task::Table::Name +
-        "(type, mrl, file_type, parent_folder_id, link_to_id, link_to_type, "
-            "link_extra)"
-            "VALUES(?, ?, ?, ?, 0, 0, 0)";
-    if ( insert( ml, self, req, Type::Creation, self->mrl(), fileType,
-                 parentFolderId ) == false )
+    auto mrl = fileFs->mrl();
+    std::shared_ptr<Task> self = std::make_shared<Task>( ml, std::move( mrl ),
+        std::move( fileFs ), std::move( parentFolder ), std::move( parentFolderFs ),
+        fileType );
+    const std::string req = "INSERT INTO " + Table::Name +
+        "(attempts_left, type, mrl, file_type, parent_folder_id, link_to_id, link_to_type, "
+            "link_extra, link_to_mrl)"
+            "VALUES(?, ?, ?, ?, ?, 0, 0, 0, '')";
+    if ( insert( ml, self, req, Settings::MaxTaskAttempts, Type::Creation,
+                 self->mrl(), fileType, parentFolderId ) == false )
         return nullptr;
-    auto parser = ml->getParser();
-    if ( parser != nullptr )
-        parser->parse( self );
     return self;
 }
 
@@ -535,55 +749,104 @@ Task::createRefreshTask( MediaLibraryPtr ml, std::shared_ptr<File> file,
     auto self = std::make_shared<Task>( ml, std::move( file ), std::move( fileFs ),
                                         std::move( parentFolder ),
                                         std::move( parentFolderFs ) );
-    const std::string req = "INSERT INTO " + Task::Table::Name +
-            "(type, mrl, file_type, file_id, parent_folder_id, link_to_id, "
-            "link_to_type, link_extra)"
-            "VALUES(?, ?, ?, ?, ?, 0, 0, 0)";
-    if ( insert( ml, self, req, Type::Refresh, self->mrl(), self->file()->type(),
-                 self->file()->id(), parentFolderId ) == false )
+    const std::string req = "INSERT INTO " + Table::Name +
+            "(attempts_left, type, mrl, file_type, file_id, parent_folder_id, link_to_id, "
+            "link_to_type, link_extra, link_to_mrl)"
+            "VALUES(?, ?, ?, ?, ?, ?, 0, 0, 0, '')";
+    if ( insert( ml, self, req, Settings::MaxTaskAttempts, Type::Refresh,
+                 self->mrl(), self->file()->type(), self->file()->id(),
+                 parentFolderId ) == false )
         return nullptr;
-    auto parser = ml->getParser();
-    if ( parser != nullptr )
-        parser->parse( self );
     return self;
 }
 
+std::shared_ptr<Task>
+Task::createMediaRefreshTask( MediaLibraryPtr ml, const Media& media )
+{
+    const auto& files = media.files();
+    auto mainFileIt = std::find_if( cbegin( files ), cend( files ),
+                                  []( const std::shared_ptr<IFile>& f ) {
+        return f->isMain();
+    });
+    if ( mainFileIt == cend( files ) )
+        return nullptr;
+    auto mainFile = std::static_pointer_cast<File>( *mainFileIt );
+    auto mrl = mainFile->mrl();
+    auto fsFactory = ml->fsFactoryForMrl( mrl );
+    if ( fsFactory == nullptr )
+        return nullptr;
+    auto folder = Folder::fetch( ml, mainFile->folderId() );
+    if ( folder == nullptr )
+        return nullptr;
+    auto folderMrl = utils::file::directory( mrl );
+    std::shared_ptr<fs::IDirectory> folderFs;
+    std::shared_ptr<fs::IFile> fileFs;
+    try
+    {
+        folderFs = fsFactory->createDirectory( folderMrl );
+        fileFs = folderFs->file( mrl );
+    }
+    catch ( const fs::errors::Exception& ex )
+    {
+        LOG_INFO( "Failed to create a media refresh task: ", ex.what() );
+        return nullptr;
+    }
+    assert( fileFs != nullptr );
+    return createRefreshTask( ml, std::move( mainFile ), std::move( fileFs ),
+                              std::move( folder ), std::move( folderFs ) );
+}
+
 std::shared_ptr<Task> Task::createLinkTask( MediaLibraryPtr ml, std::string mrl,
                                             int64_t linkToId, Task::LinkType linkToType,
                                             int64_t linkToExtra )
 {
     auto self = std::make_shared<Task>( ml, std::move( mrl ), linkToId, linkToType,
                                         linkToExtra );
-    const std::string req = "INSERT INTO " + Task::Table::Name +
-            "(type, mrl, file_type, file_id, parent_folder_id, link_to_id,"
-            "link_to_type, link_extra) VALUES(?, ?, ?, ?, ?, ?, ?, ?)";
-    if ( insert( ml, self, req, Type::Link, self->mrl(), IFile::Type::Unknown,
-                 nullptr, nullptr, linkToId, linkToType, linkToExtra ) == false )
+    const std::string req = "INSERT INTO " + Table::Name +
+            "(attempts_left, type, mrl, file_type, file_id, parent_folder_id, link_to_id,"
+            "link_to_type, link_extra, link_to_mrl) VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, '')";
+    if ( insert( ml, self, req, Settings::MaxLinkTaskAttempts, Type::Link, self->mrl(),
+                 IFile::Type::Unknown, nullptr, nullptr, linkToId, linkToType,
+                 linkToExtra ) == false )
         return nullptr;
-    auto parser = ml->getParser();
-    if ( parser != nullptr )
-        parser->parse( self );
     return self;
 }
 
-std::shared_ptr<Task> Task::createRestoreTask( MediaLibraryPtr ml, std::string mrl )
+std::shared_ptr<Task> Task::createLinkTask( MediaLibraryPtr ml, std::string mrl,
+                                            IFile::Type fileType,
+                                            std::string linkToMrl,
+                                            LinkType linkToType,
+                                            int64_t linkToExtra )
 {
-    auto self = std::make_shared<Task>( ml, std::move( mrl ) );
+    auto self = std::make_shared<Task>( ml, std::move( mrl ), fileType,
+                                        std::move( linkToMrl ), linkToType,
+                                        linkToExtra );
     const std::string req = "INSERT INTO " + Table::Name +
-            "(type, mrl, file_type, link_to_id, link_to_type, link_extra) "
-            "VALUES(?, ?, ?, 0, 0, 0)";
-    if ( insert( ml, self, req, Type::Restore, self->mrl(),
-                 IFile::Type::Unknown ) == false )
+            "(attempts_left, type, mrl, file_type, link_to_id, link_to_type, link_extra, link_to_mrl) "
+            "VALUES(?, ?, ?, ?, 0, ?, ?, ?)";
+    if ( insert( ml, self, req, Settings::MaxLinkTaskAttempts, Type::Link,
+                 self->mrl(), fileType, linkToType, linkToExtra,
+                 self->linkToMrl() ) == false )
+        return nullptr;
+    return self;
+}
+
+std::shared_ptr<Task> Task::createRestoreTask( MediaLibraryPtr ml, std::string mrl,
+                                               IFile::Type fileType )
+{
+    auto self = std::make_shared<Task>( ml, std::move( mrl ), fileType );
+    const std::string req = "INSERT INTO " + Table::Name +
+            "(attempts_left, type, mrl, file_type, link_to_id, link_to_type, link_extra, link_to_mrl) "
+            "VALUES(?, ?, ?, ?, 0, 0, 0, '')";
+    if ( insert( ml, self, req, Settings::MaxTaskAttempts, Type::Restore,
+                 self->mrl(), fileType ) == false )
         return nullptr;
-    auto parser = ml->getParser();
-    if ( parser != nullptr )
-        parser->parse( self );
     return self;
 }
 
 bool Task::removePlaylistContentTasks( MediaLibraryPtr ml, int64_t playlistId )
 {
-    const std::string req = "DELETE FROM " + Task::Table::Name + " "
+    const std::string req = "DELETE FROM " + Table::Name + " "
             "WHERE type = ? AND link_to_type = ? AND link_to_id = ?";
     return sqlite::Tools::executeDelete( ml->getConn(), req, Task::Type::Link,
                                          LinkType::Playlist, playlistId );
@@ -591,39 +854,22 @@ bool Task::removePlaylistContentTasks( MediaLibraryPtr ml, int64_t playlistId )
 
 bool Task::removePlaylistContentTasks( MediaLibraryPtr ml )
 {
-    const std::string req = "DELETE FROM " + Task::Table::Name + " "
+    const std::string req = "DELETE FROM " + Table::Name + " "
             "WHERE type = ? AND link_to_type = ?";
     return sqlite::Tools::executeDelete( ml->getConn(), req, Task::Type::Link,
                                          LinkType::Playlist );
 }
 
-/*
- * This is used only by the 11 -> 12 migration, and is refering to an old
- * DB model on purpose
- */
-bool Task::recoverUnscannedFiles( MediaLibraryPtr ml )
-{
-    static const std::string req = "INSERT INTO " + Task::Table::Name +
-            "(file_id, parent_folder_id)"
-            " SELECT id_file, folder_id FROM " + File::Table::Name +
-            " f LEFT JOIN " + Task::Table::Name + " t"
-            " ON t.file_id = f.id_file WHERE t.file_id IS NULL"
-            " AND f.folder_id IS NOT NULL";
-    return sqlite::Tools::executeInsert( ml->getConn(), req );
-}
-
 std::string Task::meta( Task::Metadata type ) const
 {
-    auto it = m_metadata.find( type );
-    if ( it == end( m_metadata ) )
-        return std::string{};
-    return it->second;
+    return m_metadata[static_cast<std::underlying_type_t<Task::Metadata>>(type)];
 }
 
 void Task::setMeta( Task::Metadata type, std::string value )
 {
-    utils::str::trim( value );
-    m_metadata[type] = std::move( value );
+    assert( type < Task::Metadata::NbValues );
+    m_metadata[static_cast<std::underlying_type_t<Task::Metadata>>(type)] =
+            utils::str::trim( std::move( value ) );
 }
 
 const std::string& Task::mrl() const
@@ -636,20 +882,36 @@ IFile::Type Task::fileType() const
     return m_fileType;
 }
 
-size_t Task::nbSubItems() const
+size_t Task::nbLinkedItems() const
 {
-    return m_subItems.size();
+    return m_linkedItems.size();
 }
 
-const IItem& Task::subItem( unsigned int index ) const
+const IItem& Task::linkedItem( unsigned int index ) const
 {
-    return m_subItems[index];
+    return m_linkedItems[index];
 }
 
-IItem& Task::createSubItem( std::string mrl, unsigned int playlistIndex )
+IItem& Task::createLinkedItem( std::string mrl, IFile::Type itemType,
+                               int64_t linkExtra )
 {
-    m_subItems.emplace_back( std::move( mrl ), IFile::Type::Main, playlistIndex );
-    return m_subItems.back();
+    LinkType linkType;
+    switch ( m_fileType )
+    {
+        case IFile::Type::Main:
+        case IFile::Type::Disc:
+            linkType = LinkType::Media;
+            break;
+        case IFile::Type::Playlist:
+            linkType = LinkType::Playlist;
+            break;
+        default:
+            throw std::runtime_error{ "Can't create a linked item for this item"
+                                      " type" };
+    }
+    m_linkedItems.emplace_back( m_ml, std::move( mrl ), itemType, m_mrl,
+                                linkType, linkExtra );
+    return m_linkedItems.back();
 }
 
 int64_t Task::duration() const
@@ -687,6 +949,11 @@ FilePtr Task::file()
     return m_file;
 }
 
+int64_t Task::fileId() const
+{
+    return m_fileId;
+}
+
 bool Task::setFile( FilePtr file )
 {
     auto fileId = file->id();
@@ -694,7 +961,7 @@ bool Task::setFile( FilePtr file )
         return true ;
     assert( m_fileId == 0 );
     assert( fileId != 0 );
-    static const std::string req = "UPDATE " + Task::Table::Name + " SET "
+    static const std::string req = "UPDATE " + Table::Name + " SET "
             "file_id = ? WHERE id_task = ?";
     if ( sqlite::Tools::executeUpdate( m_ml->getConn(), req, fileId, m_id ) == false )
         return false;
@@ -708,7 +975,7 @@ FolderPtr Task::parentFolder()
     return m_parentFolder;
 }
 
-std::shared_ptr<fs::IFile> Task::fileFs()
+std::shared_ptr<fs::IFile> Task::fileFs() const
 {
     return m_fileFs;
 }
@@ -748,6 +1015,32 @@ int64_t Task::linkExtra() const
     return m_linkExtra;
 }
 
+const std::string& Task::linkToMrl() const
+{
+    return m_linkToMrl;
+}
+
+bool Task::needEntityRestoration() const
+{
+    if ( isLinkTask() == true || isRestore() == true )
+        return false;
+    return m_parentFolderFs == nullptr ||
+           m_fileFs == nullptr ||
+           m_parentFolder == nullptr ||
+           ( m_file == nullptr && m_fileId != 0 );
+}
+
+const std::vector<std::shared_ptr<IEmbeddedThumbnail>>&
+Task::embeddedThumbnails() const
+{
+    return m_embeddedThumbnails;
+}
+
+void Task::addEmbeddedThumbnail( std::shared_ptr<IEmbeddedThumbnail> t )
+{
+    m_embeddedThumbnails.push_back( std::move( t ) );
+}
+
 }
 
 }
