@@ -26,39 +26,60 @@
 
 #include "Tester.h"
 
+#include "common/util.h"
 #include "parser/Parser.h"
 #include "Thumbnail.h"
 #include "File.h"
 #include "utils/Filename.h"
+#include "utils/Url.h"
+#include "utils/Directory.h"
 #include "factory/DeviceListerFactory.h"
+#include "medialibrary/filesystem/IFile.h"
+#include "medialibrary/filesystem/IDirectory.h"
+#include "medialibrary/IShow.h"
+#include "medialibrary/IShowEpisode.h"
+#include "medialibrary/IMediaGroup.h"
+#include "utils/Directory.h"
+#include "filesystem/libvlc/FileSystemFactory.h"
 
 #include <algorithm>
 
-extern bool Verbose;
-extern bool ExtraVerbose;
+const std::string Tests::Directory = SRC_DIR "/test/samples/";
 
 MockCallback::MockCallback()
+    : m_thumbnailDone( false )
+    , m_thumbnailSuccess( false )
+    , m_parserDone( false )
+    , m_discoveryCompleted( false )
+    , m_removalCompleted( false )
+    , m_nbEntryPointsRemovalExpected( 0 )
 {
-    // Start locked. The locked will be released when waiting for parsing to be completed
-    m_discoveryCompleted = false;
-    m_done = false;
 }
 
-std::unique_lock<compat::Mutex> MockCallback::lock()
+void MockCallback::waitForParsingComplete()
 {
-    return std::unique_lock<compat::Mutex>{ m_parsingMutex };
-}
-
-bool MockCallback::waitForParsingComplete( std::unique_lock<compat::Mutex>& lock )
-{
-    m_done = false;
-    m_discoveryCompleted = false;
+    std::unique_lock<compat::Mutex> lock{ m_parsingMutex };
     // Wait for a while, generating snapshots can be heavy...
-    return m_parsingCompleteVar.wait_for( lock, std::chrono::seconds{ 20 }, [this]() {
-        return m_done;
+    m_parsingCompleteVar.wait( lock, [this]() {
+        return m_parserDone && m_discoveryCompleted;
     });
 }
 
+bool MockCallback::waitForRemovalComplete()
+{
+    std::unique_lock<compat::Mutex> lock{ m_parsingMutex };
+    return m_parsingCompleteVar.wait_for( lock, std::chrono::seconds{ 20 }, [this]() {
+        return m_removalCompleted;
+    });
+}
+
+void MockCallback::reinit()
+{
+    std::lock_guard<compat::Mutex> lock( m_parsingMutex );
+    m_discoveryCompleted = false;
+    m_parserDone = false;
+}
+
 void MockCallback::prepareWaitForThumbnail( MediaPtr media )
 {
     m_thumbnailMutex.lock();
@@ -77,24 +98,24 @@ bool MockCallback::waitForThumbnail()
     return m_thumbnailSuccess;
 }
 
-void MockCallback::onDiscoveryCompleted(const std::string& entryPoint, bool )
+
+void MockCallback::onDiscoveryStarted()
+{
+    std::lock_guard<compat::Mutex> lock( m_parsingMutex );
+    m_discoveryCompleted = false;
+}
+
+void MockCallback::onDiscoveryCompleted()
 {
-    if ( entryPoint.empty() == true )
-        return;
     std::lock_guard<compat::Mutex> lock( m_parsingMutex );
     m_discoveryCompleted = true;
 }
 
-void MockCallback::onParsingStatsUpdated(uint32_t percent)
+void MockCallback::onParsingStatsUpdated( uint32_t done, uint32_t scheduled )
 {
-    if ( percent == 100 )
-    {
-        std::lock_guard<compat::Mutex> lock( m_parsingMutex );
-        if ( m_discoveryCompleted == false )
-            return;
-        m_done = true;
-        m_parsingCompleteVar.notify_all();
-    }
+    std::lock_guard<compat::Mutex> lock( m_parsingMutex );
+
+    m_parserDone = done == scheduled;
 }
 
 void MockCallback::onMediaThumbnailReady( MediaPtr media, ThumbnailSizeType,
@@ -109,15 +130,25 @@ void MockCallback::onMediaThumbnailReady( MediaPtr media, ThumbnailSizeType,
     m_thumbnailCond.notify_all();
 }
 
-MockResumeCallback::MockResumeCallback()
+void MockCallback::onEntryPointRemoved( const std::string& entryPoint, bool )
 {
+    assert( entryPoint.empty() == false );
+    std::lock_guard<compat::Mutex> lock( m_parsingMutex );
+    assert( m_nbEntryPointsRemovalExpected > 0 );
+    if ( --m_nbEntryPointsRemovalExpected > 0 )
+        return;
+    m_removalCompleted = true;
 }
 
-void MockResumeCallback::onDiscoveryCompleted( const std::string& entryPoint, bool )
+void MockCallback::onBackgroundTasksIdleChanged( bool idle )
 {
-    if ( entryPoint.empty() == true )
+    if ( idle == false )
         return;
+    m_parsingCompleteVar.notify_all();
+}
 
+void MockResumeCallback::onDiscoveryCompleted()
+{
     std::lock_guard<compat::Mutex> lock( m_parsingMutex );
     m_discoveryCompleted = true;
     m_discoveryCompletedVar.notify_all();
@@ -125,54 +156,85 @@ void MockResumeCallback::onDiscoveryCompleted( const std::string& entryPoint, bo
 
 void MockResumeCallback::reinit()
 {
+    std::lock_guard<compat::Mutex> lock( m_parsingMutex );
     m_discoveryCompleted = true;
-    m_done = false;
+    m_parserDone = false;
 }
 
-bool MockResumeCallback::waitForDiscoveryComplete( std::unique_lock<compat::Mutex>& lock )
+void MockResumeCallback::waitForDiscoveryComplete()
 {
-    assert( m_discoveryCompleted == false );
-    // Wait for a while, generating snapshots can be heavy...
-    return m_discoveryCompletedVar.wait_for( lock, std::chrono::seconds{ 20 }, [this]() {
+    std::unique_lock<compat::Mutex> lock{ m_parsingMutex };
+    m_discoveryCompletedVar.wait( lock, [this]() {
         return m_discoveryCompleted;
     });
 }
 
-bool MockResumeCallback::waitForParsingComplete( std::unique_lock<compat::Mutex>& lock )
+void MockResumeCallback::waitForParsingComplete()
 {
+    std::unique_lock<compat::Mutex> lock{ m_parsingMutex };
     // Reimplement without checking for discovery complete. This class is meant to be used
     // in 2 steps: waiting for discovery completed, then for parsing completed
     assert( m_discoveryCompleted == true );
-    m_done = false;
-    // Wait for a while, generating snapshots can be heavy...
-    return m_parsingCompleteVar.wait_for( lock, std::chrono::seconds{ 20 }, [this]() {
-        return m_done;
+    m_parsingCompleteVar.wait( lock, [this]() {
+        return m_parserDone;
     });
 }
 
-void Tests::SetUp()
+void Tests::InitTestCase( const std::string& testName )
+{
+    auto casePath = Directory + "testcases/" + testName + ".json";
+    std::unique_ptr<FILE, int(*)(FILE*)> f( fopen( casePath.c_str(), "rb" ), &fclose );
+    ASSERT_NE( nullptr, f );
+    char buff[65536];
+    auto ret = fread( buff, sizeof(buff[0]), sizeof(buff), f.get() );
+    ASSERT_NE( 0u, ret );
+    buff[ret] = 0;
+    doc.Parse( buff );
+
+    if ( doc.HasMember( "banned" ) == true )
+    {
+        const auto& banned = doc["banned"];
+        for ( auto i = 0u; i < banned.Size(); ++i )
+        {
+            auto bannedDir = Directory + "samples/" + banned[i].GetString();
+            ASSERT_TRUE( utils::fs::isDirectory( bannedDir ) );
+            bannedDir = utils::fs::toAbsolute( bannedDir );
+            m_ml->banFolder( utils::file::toMrl( bannedDir ) );
+        }
+    }
+
+    ASSERT_TRUE( doc.HasMember( "input" ) );
+    input = doc["input"];
+    for ( auto i = 0u; i < input.Size(); ++i )
+    {
+        // Quick and dirty check to ensure we're discovering something that exists
+        auto samplesDir = Directory + "samples/" + input[i].GetString();
+        ASSERT_TRUE( utils::fs::isDirectory( samplesDir ) );
+        samplesDir = utils::fs::toAbsolute( samplesDir );
+
+        m_ml->discover( utils::file::toMrl( samplesDir ) );
+    }
+}
+
+void Tests::SetUp( const std::string& testSuite, const std::string& testName )
 {
-    unlink("test.db");
     InitializeCallback();
-    InitializeMediaLibrary();
-    if ( ExtraVerbose == true )
-        m_ml->setVerbosity( LogLevel::Debug );
-    else if ( Verbose == true )
-        m_ml->setVerbosity( LogLevel::Info );
+    m_testDir = getTempPath( testSuite + "." + testName );
+    auto dbPath = m_testDir + "/test.db";
+    InitializeMediaLibrary( dbPath, m_testDir );
+    m_ml->setVerbosity( LogLevel::Debug );
 
-#ifndef _WIN32
-    auto mlDir = "/tmp/ml_folder/";
-#else
-    // This assumes wine for now
-    auto mlDir = "Z:\\tmp\\ml_folder\\";
-#endif
-
-    if ( std::get<1>( GetParam() ) == true )
-        m_ml->setDeviceLister( std::make_shared<ForceRemovableStorareDeviceLister>() );
-
-    auto res = m_ml->initialize( "test.db", mlDir, m_cb.get() );
+    auto res = m_ml->initialize( m_cb.get() );
     ASSERT_EQ( InitializeResult::Success, res );
-    ASSERT_TRUE( m_ml->start() );
+
+    InitTestCase( testName );
+}
+
+void Tests::TearDown()
+{
+    /* Ensure we are closing our database connection before we try to delete it */
+    m_ml.reset();
+    ASSERT_TRUE( utils::fs::rmdir( m_testDir ) );
 }
 
 void Tests::InitializeCallback()
@@ -180,12 +242,34 @@ void Tests::InitializeCallback()
     m_cb.reset( new MockCallback );
 }
 
-void Tests::InitializeMediaLibrary()
+namespace
 {
-    m_ml.reset( NewMediaLibrary() );
+class MediaLibraryTester : public MediaLibrary
+{
+    using MediaLibrary::MediaLibrary;
+    virtual ~MediaLibraryTester()
+    {
+        stopBackgroundJobs();
+    }
+    virtual void onDbConnectionReady( sqlite::Connection* dbConn ) override
+    {
+        sqlite::Connection::DisableForeignKeyContext ctx{ m_dbConnection.get() };
+        auto t = m_dbConnection->newTransaction();
+        deleteAllTables( dbConn );
+        t->commit();
+        m_dbConnection->flushAll();
+    }
+};
 }
 
-void Tests::runChecks(const rapidjson::Document& doc)
+void Tests::InitializeMediaLibrary( const std::string& dbPath,
+                                    const std::string& mlFolderDir )
+
+{
+    m_ml.reset( new MediaLibraryTester{ dbPath, mlFolderDir } );
+}
+
+void Tests::runChecks()
 {
     if ( doc.HasMember( "expected" ) == false )
     {
@@ -212,21 +296,23 @@ void Tests::runChecks(const rapidjson::Document& doc)
     }
     if ( expected.HasMember( "nbPlaylists" ) == true )
     {
-        const auto playlists = m_ml->playlists( nullptr )->all();
+        const auto playlists = m_ml->playlists( PlaylistType::All, nullptr )->all();
         ASSERT_EQ( expected["nbPlaylists"].GetUint(), playlists.size() );
     }
     if ( expected.HasMember( "playlists" ) == true )
     {
-      checkPlaylists( expected["playlists"], m_ml->playlists( nullptr )->all() );
+        checkPlaylists( expected["playlists"], m_ml->playlists( PlaylistType::All,
+                                                                nullptr )->all() );
     }
     if ( expected.HasMember( "artists" ) )
     {
-        checkArtists( expected["artists"], m_ml->artists( true, nullptr )->all() );
+        checkArtists( expected["artists"], m_ml->artists( ArtistIncluded::All, nullptr )->all() );
     }
     if ( expected.HasMember( "nbThumbnails" ) )
     {
+        auto ml = static_cast<MediaLibrary*>( m_ml.get() );
+        OPEN_READ_CONTEXT( ctx, ml->getConn() );
         sqlite::Statement stmt{
-            static_cast<MediaLibrary*>( m_ml.get() )->getConn()->handle(),
             "SELECT COUNT(*) FROM " + Thumbnail::Table::Name
         };
         uint32_t nbThumbnails;
@@ -235,6 +321,15 @@ void Tests::runChecks(const rapidjson::Document& doc)
         row >> nbThumbnails;
         ASSERT_EQ( expected["nbThumbnails"].GetUint(), nbThumbnails );
     }
+    if ( expected.HasMember( "shows" ) == true )
+    {
+        checkShows( expected["shows"], m_ml->shows( nullptr )->all() );
+    }
+    if ( expected.HasMember( "mediaGroups" ) == true )
+    {
+        checkMediaGroups( expected["mediaGroups"],
+                m_ml->mediaGroups( IMedia::Type::Unknown, nullptr )->all() );
+    }
 }
 
 void Tests::checkVideoTracks( const rapidjson::Value& expectedTracks, const std::vector<VideoTrackPtr>& tracks )
@@ -250,7 +345,8 @@ void Tests::checkVideoTracks( const rapidjson::Value& expectedTracks, const std:
         ASSERT_TRUE( expectedTrack.IsObject() );
         if ( expectedTrack.HasMember( "codec" ) )
         {
-            ASSERT_STRCASEEQ( expectedTrack["codec"].GetString(), track->codec().c_str() );
+            ASSERT_EQ( strcasecmp( expectedTrack["codec"].GetString(),
+                                   track->codec().c_str() ), 0 );
         }
         if ( expectedTrack.HasMember( "width" ) )
         {
@@ -278,7 +374,8 @@ void Tests::checkAudioTracks(const rapidjson::Value& expectedTracks, const std::
         ASSERT_TRUE( expectedTrack.IsObject() );
         if ( expectedTrack.HasMember( "codec" ) )
         {
-            ASSERT_STRCASEEQ( expectedTrack["codec"].GetString(), track->codec().c_str() );
+            ASSERT_EQ( strcasecmp( expectedTrack["codec"].GetString(),
+                                   track->codec().c_str() ), 0 );
         }
         if ( expectedTrack.HasMember( "sampleRate" ) )
         {
@@ -307,15 +404,41 @@ void Tests::checkSubtitleTracks( const rapidjson::Value& expectedTracks,
         ASSERT_TRUE( expectedTrack.IsObject() );
         if ( expectedTrack.HasMember( "codec" ) )
         {
-            ASSERT_STRCASEEQ( expectedTrack["codec"].GetString(), track->codec().c_str() );
+            ASSERT_EQ( strcasecmp( expectedTrack["codec"].GetString(),
+                                   track->codec().c_str() ), 0 );
         }
         if ( expectedTrack.HasMember( "encoding" ) )
         {
-            ASSERT_STRCASEEQ( expectedTrack["encoding"].GetString(), track->encoding().c_str() );
+            ASSERT_EQ( strcasecmp( expectedTrack["encoding"].GetString(),
+                                   track->encoding().c_str() ), 0 );
         }
     }
 }
 
+void Tests::checkMediaFiles( const IMedia *media, const rapidjson::Value& expectedFiles )
+{
+    ASSERT_TRUE( expectedFiles.IsArray() );
+    auto files = media->files();
+    ASSERT_EQ( expectedFiles.Size(), files.size() );
+    for ( auto i = 0u; i < expectedFiles.Size(); ++i )
+    {
+        const auto& expectedFile = expectedFiles[i];
+        ASSERT_TRUE( expectedFile.HasMember( "filename" ) );
+        auto it = std::find_if( begin( files ), end( files ), [&expectedFile]( FilePtr f ) {
+            return utils::file::fileName( f->mrl() ) == expectedFile["filename"].GetString();
+        });
+        ASSERT_TRUE( it != end( files ) );
+
+        if ( expectedFile.HasMember( "type" ) )
+        {
+            auto expectedType = expectedFile["type"].GetInt();
+            ASSERT_EQ( expectedType,
+                       static_cast<std::underlying_type_t<IFile::Type>>( (*it)->type() ) );
+        }
+        files.erase( it );
+}
+}
+
 void Tests::checkMedias(const rapidjson::Value& expectedMedias)
 {
     ASSERT_TRUE( expectedMedias.IsArray() );
@@ -330,7 +453,7 @@ void Tests::checkMedias(const rapidjson::Value& expectedMedias)
         auto it = std::find_if( begin( medias ), end( medias ), [expectedTitle](const MediaPtr& m) {
             return strcasecmp( expectedTitle, m->title().c_str() ) == 0;
         });
-        ASSERT_NE( end( medias ), it );
+        ASSERT_TRUE( end( medias ) != it );
         const auto media = *it;
         medias.erase( it );
         if ( expectedMedia.HasMember( "nbVideoTracks" ) || expectedMedia.HasMember( "videoTracks" ) )
@@ -355,6 +478,16 @@ void Tests::checkMedias(const rapidjson::Value& expectedMedias)
             if ( expectedMedia.HasMember( "audioTracks" ) )
                 checkAudioTracks( expectedMedia[ "audioTracks" ], audioTracks );
         }
+        if ( expectedMedia.HasMember( "nbSubtitleTracks" ) || expectedMedia.HasMember( "subtitleTracks" ) )
+        {
+            auto subtitleTracks = media->subtitleTracks()->all();
+            if ( expectedMedia.HasMember( "nbSubtitleTracks" ) )
+            {
+                ASSERT_EQ( expectedMedia[ "nbSubtitleTracks" ].GetUint(), subtitleTracks.size() );
+            }
+            if ( expectedMedia.HasMember( "subtitleTracks" ) )
+                checkSubtitleTracks( expectedMedia[ "subtitleTracks" ], subtitleTracks );
+        }
         if ( expectedMedia.HasMember( "snapshotExpected" ) == true )
         {
             auto snapshotExpected = expectedMedia["snapshotExpected"].GetBool();
@@ -367,6 +500,10 @@ void Tests::checkMedias(const rapidjson::Value& expectedMedias)
             }
             ASSERT_EQ( !snapshotExpected, media->thumbnailMrl( ThumbnailSizeType::Thumbnail ).empty() );
         }
+        if ( expectedMedia.HasMember( "files" ) )
+        {
+            checkMediaFiles( media.get(), expectedMedia["files"] );
+        }
     }
 }
 
@@ -381,11 +518,11 @@ void Tests::checkPlaylists( const rapidjson::Value& expectedPlaylists, std::vect
         auto it = std::find_if( begin( playlists ), end( playlists ), [expectedName](const PlaylistPtr& p) {
             return strcasecmp( expectedName, p->name().c_str() ) == 0;
         });
-        ASSERT_NE( end( playlists ), it );
+        ASSERT_TRUE( end( playlists ) != it );
 
         const auto playlist = *it;
         playlists.erase( it );
-        const auto& items = playlist->media()->all();
+        const auto& items = playlist->media( nullptr )->all();
 
         ASSERT_TRUE( playlist->isReadOnly() );
         ASSERT_FALSE( playlist->mrl().empty() );
@@ -394,20 +531,40 @@ void Tests::checkPlaylists( const rapidjson::Value& expectedPlaylists, std::vect
         {
             ASSERT_EQ( expectedPlaylist["nbItems"].GetUint(), items.size() );
         }
+        if ( expectedPlaylist.HasMember( "nbAudio" ) )
+        {
+            ASSERT_EQ( expectedPlaylist["nbAudio"].GetUint(), playlist->nbAudio() );
+        }
+        if ( expectedPlaylist.HasMember( "nbDurationUnknown" ) )
+        {
+            ASSERT_EQ( expectedPlaylist["nbDurationUnknown"].GetUint(),
+                    playlist->nbDurationUnknown() );
+        }
 
         if ( expectedPlaylist.HasMember( "items" ) )
         {
             ASSERT_TRUE( expectedPlaylist["items"].IsArray() );
             ASSERT_EQ( items.size(), expectedPlaylist["items"].Size() );
-            for ( auto i = 0u; i < items.size(); ++i )
+            for ( auto j = 0u; j < items.size(); ++j )
             {
-                if ( expectedPlaylist["items"][i].HasMember( "index" ) )
+                if ( expectedPlaylist["items"][j].HasMember( "index" ) )
                 {
-                    ASSERT_EQ( expectedPlaylist["items"][i]["index"].GetUint(), i );
+                    ASSERT_EQ( expectedPlaylist["items"][j]["index"].GetUint(), j );
                 }
-                if ( expectedPlaylist["items"][i].HasMember( "title" ) )
+                if ( expectedPlaylist["items"][j].HasMember( "title" ) )
                 {
-                    ASSERT_EQ( expectedPlaylist["items"][i]["title"].GetString(), items[i]->title() );
+                    ASSERT_EQ( expectedPlaylist["items"][j]["title"].GetString(), items[j]->title() );
+                }
+                if ( expectedPlaylist["items"][j].HasMember( "mrl" ) )
+                {
+                    auto files = items[j]->files();
+                    auto mainFileIt = std::find_if( cbegin( files ), cend( files ),
+                        [](const FilePtr& f ) {
+                            return f->isMain() == true;
+                    });
+                    ASSERT_TRUE( mainFileIt != cend( files ) );
+                    ASSERT_EQ( expectedPlaylist["items"][j]["mrl"].GetString(),
+                            (*mainFileIt)->mrl() );
                 }
             }
         }
@@ -450,6 +607,14 @@ void Tests::checkAlbums( const rapidjson::Value& expectedAlbums, std::vector<Alb
                         return false;
                 }
             }
+            if ( expectedAlbum.HasMember( "hasArtwork" ) )
+            {
+                if ( expectedAlbum["hasArtwork"].GetBool() ==
+                     a->thumbnailMrl( ThumbnailSizeType::Thumbnail ).empty() ||
+                     a->thumbnailMrl( ThumbnailSizeType::Thumbnail )
+                        .compare( 0, 13, "attachment://") == 0 )
+                    return false;
+            }
             if ( expectedAlbum.HasMember( "nbTracks" ) || expectedAlbum.HasMember( "tracks" ) )
             {
                 const auto tracks = a->tracks( nullptr )->all();
@@ -472,14 +637,6 @@ void Tests::checkAlbums( const rapidjson::Value& expectedAlbums, std::vector<Alb
                 if ( a->releaseYear() != releaseYear )
                     return false;
             }
-            if ( expectedAlbum.HasMember( "hasArtwork" ) )
-            {
-                if ( expectedAlbum["hasArtwork"].GetBool() ==
-                     a->thumbnailMrl( ThumbnailSizeType::Thumbnail ).empty() ||
-                     a->thumbnailMrl( ThumbnailSizeType::Thumbnail )
-                        .compare( 0, 13, "attachment://") == 0 )
-                    return false;
-            }
             if ( expectedAlbum.HasMember( "nbDiscs" ) )
             {
                 const auto nbDiscs = expectedAlbum["nbDiscs"].GetUint();
@@ -488,7 +645,7 @@ void Tests::checkAlbums( const rapidjson::Value& expectedAlbums, std::vector<Alb
             }
             return true;
         });
-        ASSERT_NE( end( albums ), it );
+        ASSERT_TRUE( end( albums ) != it );
         albums.erase( it );
     }
 }
@@ -539,7 +696,7 @@ void Tests::checkArtists(const rapidjson::Value& expectedArtists, std::vector<Ar
             }
             return true;
         });
-        ASSERT_NE( it, end( artists ) );
+        ASSERT_TRUE( it != end( artists ) );
     }
 }
 
@@ -558,15 +715,15 @@ void Tests::checkAlbumTracks( const IAlbum* album, const std::vector<MediaPtr>&
         if ( it == end( tracks ) )
             return ;
         const auto track = *it;
-        const auto albumTrack = track->albumTrack();
+        ASSERT_NE( nullptr, track );
         if ( expectedTrack.HasMember( "number" ) )
         {
-            if ( expectedTrack["number"].GetUint() != albumTrack->trackNumber() )
+            if ( expectedTrack["number"].GetUint() != track->trackNumber() )
                 return ;
         }
         if ( expectedTrack.HasMember( "artist" ) )
         {
-            auto artist = albumTrack->artist();
+            auto artist = track->artist();
             if ( artist == nullptr )
                 return ;
             if ( strlen( expectedTrack["artist"].GetString() ) == 0 &&
@@ -577,7 +734,7 @@ void Tests::checkAlbumTracks( const IAlbum* album, const std::vector<MediaPtr>&
         }
         if ( expectedTrack.HasMember( "genre" ) )
         {
-            auto genre = albumTrack->genre();
+            auto genre = track->genre();
             if ( genre == nullptr || strcasecmp( expectedTrack["genre"].GetString(), genre->name().c_str() ) != 0 )
                 return ;
         }
@@ -588,7 +745,7 @@ void Tests::checkAlbumTracks( const IAlbum* album, const std::vector<MediaPtr>&
         }
         if ( expectedTrack.HasMember( "cd" ) )
         {
-            if ( expectedTrack["cd"].GetUint() != albumTrack->discNumber() )
+            if ( expectedTrack["cd"].GetUint() != track->discNumber() )
                 return;
         }
         if ( expectedTrack.HasMember( "hasArtwork" ) )
@@ -600,16 +757,105 @@ void Tests::checkAlbumTracks( const IAlbum* album, const std::vector<MediaPtr>&
         }
         // Always check if the album link is correct. This isn't part of finding the proper album, so just fail hard
         // if the check fails.
-        const auto trackAlbum = albumTrack->album();
+        const auto trackAlbum = track->album();
         ASSERT_NE( nullptr, trackAlbum );
         ASSERT_EQ( album->id(), trackAlbum->id() );
     }
     found = true;
 }
 
-void ResumeTests::InitializeMediaLibrary()
+void Tests::checkShows(const rapidjson::Value& expectedShows, std::vector<ShowPtr> shows)
 {
-    m_ml.reset( new MediaLibraryResumeTest );
+    for ( auto i = 0u; i < expectedShows.Size(); ++i )
+    {
+        auto& expectedShow = expectedShows[i];
+        ASSERT_TRUE( expectedShow.HasMember( "name" ) );
+        auto showName = expectedShow["name"].GetString();
+        auto showIt = std::find_if( cbegin( shows ), cend( shows ),
+                                  [showName]( const ShowPtr& s ) {
+            if ( strlen( showName ) == 0 )
+                return s->id() == UnknownShowID;
+            return s->title() == showName;
+        });
+        ASSERT_TRUE( showIt != cend( shows ) );
+        auto show = *showIt;
+        if ( expectedShow.HasMember( "nbEpisodes" ) == true )
+        {
+            ASSERT_EQ( expectedShow["nbEpisodes"].GetUint(), show->nbEpisodes() );
+        }
+        if ( expectedShow.HasMember( "episodes" ) == true )
+        {
+            auto episodes = show->episodes( nullptr )->all();
+            ASSERT_FALSE( episodes.empty() );
+            checkShowEpisodes( expectedShow["episodes"], std::move( episodes ) );
+        }
+    }
+}
+
+void Tests::checkShowEpisodes( const rapidjson::Value& expectedEpisodes,
+                               std::vector<MediaPtr> episodes )
+{
+    for ( auto i = 0u; i < expectedEpisodes.Size(); ++i )
+    {
+        auto& expectedEpisode = expectedEpisodes[i];
+        ASSERT_TRUE( expectedEpisode.HasMember( "seasonId" ) );
+        ASSERT_TRUE( expectedEpisode.HasMember( "episodeId" ) );
+        auto seasonId = expectedEpisode["seasonId"].GetUint();
+        auto episodeId = expectedEpisode["episodeId"].GetUint();
+        auto episodeIt = std::find_if( cbegin( episodes ), cend( episodes ),
+                                       [seasonId, episodeId](const MediaPtr m) {
+            auto showEp = m->showEpisode();
+            if ( showEp == nullptr )
+                return false;
+            return showEp->seasonId() == seasonId && showEp->episodeId() == episodeId;
+        });
+        ASSERT_TRUE( episodeIt != cend( episodes ) );
+        auto media = *episodeIt;
+        auto showEp = media->showEpisode();
+        if ( expectedEpisode.HasMember( "title" ) == true )
+        {
+            ASSERT_EQ( expectedEpisode["title"].GetString(), showEp->title() );
+        }
+        if ( expectedEpisode.HasMember( "mediaTitle" ) == true )
+        {
+            ASSERT_EQ( expectedEpisode["mediaTitle"].GetString(), media->title() );
+        }
+    }
+}
+
+void Tests::checkMediaGroups( const rapidjson::Value &expectedMediaGroups,
+                              std::vector<MediaGroupPtr> mediaGroups)
+{
+    ASSERT_EQ( expectedMediaGroups.Size(), mediaGroups.size() );
+    for ( auto i = 0u; i < expectedMediaGroups.Size(); ++i )
+    {
+        auto& expectedGroup = expectedMediaGroups[i];
+        ASSERT_TRUE( expectedGroup.HasMember( "name" ) );
+        auto it = std::find_if( cbegin( mediaGroups ), cend( mediaGroups ),
+                                [&expectedGroup]( const MediaGroupPtr grp ) {
+            return strcasecmp( grp->name().c_str(), expectedGroup["name"].GetString() ) == 0;
+        });
+        ASSERT_TRUE( cend( mediaGroups ) != it );
+        auto group = *it;
+        if ( expectedGroup.HasMember( "nbAudio" ) )
+        {
+            ASSERT_EQ( expectedGroup["nbAudio"].GetUint(), group->nbPresentAudio() );
+        }
+        if ( expectedGroup.HasMember( "nbVideo" ) )
+        {
+            ASSERT_EQ( expectedGroup["nbVideo"].GetUint(), group->nbPresentVideo() );
+        }
+        if ( expectedGroup.HasMember( "nbUnknown" ) )
+        {
+            ASSERT_EQ( expectedGroup["nbUnknown"].GetUint(), group->nbPresentUnknown() );
+        }
+    }
+}
+
+void ResumeTests::InitializeMediaLibrary( const std::string& dbPath,
+                                          const std::string& mlFolderDir )
+{
+    m_ml.reset( new MediaLibraryResumeTest( dbPath, mlFolderDir ) );
 }
 
 void ResumeTests::InitializeCallback()
@@ -617,22 +863,38 @@ void ResumeTests::InitializeCallback()
     m_cb.reset( new MockResumeCallback );
 }
 
-void ResumeTests::MediaLibraryResumeTest::forceParserStart()
+MediaLibraryResumeTest::~MediaLibraryResumeTest()
 {
-    MediaLibrary::startParser();
+    stopBackgroundJobs();
 }
 
-bool ResumeTests::MediaLibraryResumeTest::startParser()
+void MediaLibraryResumeTest::forceParserStart()
 {
-    return true;
+    m_allowParser = true;
+    getParser()->start();
 }
 
+void MediaLibraryResumeTest::onDbConnectionReady( sqlite::Connection *dbConn )
+{
+    sqlite::Connection::DisableForeignKeyContext ctx{ m_dbConnection.get() };
+    auto t = m_dbConnection->newTransaction();
+    deleteAllTables( dbConn );
+    t->commit();
+    m_dbConnection->flushAll();
+}
+
+parser::Parser* MediaLibraryResumeTest::getParser() const
+{
+    if ( m_allowParser == false )
+        return nullptr;
+    return MediaLibrary::getParser();
+}
 
 void RefreshTests::forceRefresh()
 {
     auto ml = static_cast<MediaLibrary*>( m_ml.get() );
     auto files = medialibrary::File::fetchAll( ml );
-    for ( const auto& f : files )
+    for ( auto& f : files )
     {
         if ( f->isExternal() == true )
             continue;
@@ -648,7 +910,7 @@ void RefreshTests::forceRefresh()
                 return f->name() == fileName;
             });
         assert( fileFsIt != cend( filesFs ) );
-        ml->onUpdatedFile( f, *fileFsIt, std::move( folder ), std::move( folderFs ) );
+        ml->onUpdatedFile( std::move( f ), *fileFsIt, std::move( folder ), std::move( folderFs ) );
     }
 }
 
@@ -657,17 +919,48 @@ void RefreshTests::InitializeCallback()
     m_cb.reset( new MockResumeCallback );
 }
 
-ForceRemovableStorareDeviceLister::ForceRemovableStorareDeviceLister()
+void MockCallback::prepareForPlaylistReload()
 {
-    m_lister = factory::createDeviceLister();
-    assert( m_lister != nullptr );
+    // We need to force the discover to appear as complete, as we won't do any
+    // discovery for this test. Otherwise, we'd receive the parsing completed
+    // event and just ignore it.
+    std::lock_guard<compat::Mutex> lock{ m_parsingMutex };
+    m_discoveryCompleted = true;
+    m_parserDone = false;
 }
 
-std::vector<std::tuple<std::string, std::string, bool>>
-ForceRemovableStorareDeviceLister::devices() const
+void MockCallback::waitForPlaylistReload()
 {
-    auto devices = m_lister->devices();
-    for ( auto& d : devices )
-        std::get<2>( d ) = true;
-    return devices;
+    std::unique_lock<compat::Mutex> lock{ m_parsingMutex };
+    // Wait for a while, generating snapshots can be heavy...
+    m_parsingCompleteVar.wait( lock, [this]() {
+        return m_parserDone;
+    });
+}
+
+void MockCallback::prepareForRemoval( uint32_t nbEntryPointsRemovalExpected )
+{
+    std::lock_guard<compat::Mutex> lock{ m_parsingMutex };
+    m_nbEntryPointsRemovalExpected = nbEntryPointsRemovalExpected;
+    m_removalCompleted = false;
+}
+
+void ReplaceExternalMediaByPlaylistTests::InitTestCase( const std::string& )
+{
+    /*
+     * This test is about recovering from a playlist wrongly inserted as an
+     * external media, which can happen if the user starts an unimported
+     * playlist playback, only to discover the folder containing the playlist at
+     * a later time.
+     * After the initial playback, VLC will create an external media to save the
+     * playback states & preferences.
+     * When the playlist should be imported, there's already a file representing
+     * that playlist so it fails to get imported.
+     * See #400
+     */
+    auto playlistPath = utils::fs::toAbsolute( SRC_DIR );
+    playlistPath += "test/samples/samples/playlist/mixed_content/playlist.xspf";
+    playlistMrl = utils::file::toMrl( playlistPath );
+    m_ml->addExternalMedia( playlistMrl, -1 );
+    Tests::InitTestCase( "playlist_mixed_content" );
 }
